{-# LANGUAGE TemplateHaskell #-}
module HsProver.HsProver
where
import LambdaP.Core
import LambdaP.GenChecker
import qualified Data.Set as Set
import qualified Data.Map as Map
import Data.Map (Map)
import Data.Array.IArray
import Data.Array.Unboxed
import Data.Array.ST
import Data.Foldable
import Data.Word
import Data.Bool
import Data.Tree
import Data.Maybe

import Control.Monad
import Control.Monad.Trans.Class
import Control.Monad.State.Class
import qualified Control.Monad.Trans.State as Cmts

import Control.Lens

-- nodes of tree A have id's
data NodeA = NodeA { _idA :: Word64 }
	deriving (Ord, Eq, Show)

data State = State { _idS :: Word64, _lvl :: NodeA } 
	deriving (Ord, Eq, Show)

-- a Fast tree
data Fast x = Fast { 	_nodeF :: x, -- the value
			_parentF :: x, -- parent
--			_left :: Maybe x, _right :: Fast x, -- siblings
			_left :: (), _right :: (), -- siblings
--			REMOVED, we keep the siblings globally (for now)
			_childrenF :: Map NodeA (Fast x)} -- first child
	deriving (Ord, Eq, Show)

data CompRes = CLT | CGT | CEQ | CIN
	deriving (Ord, Eq, Show)

collectFast = (tmp (Map.empty))
	where
		tmp map f@(Fast node parent _ _ children) =
			let map'  = Map.insert node f map in -- add this node
			let map'' = Map.fold (flip tmp) map' children in -- add children 
			-- next line is unneeded: if we have any siblings, they're processed as our parent's children
			-- maybe map'' (tmp map'') children -- process sibling if any
			map''

collectNormal :: Ord x => Normal x -> Map x (Normal x)
collectNormal = (tmp (Map.empty))
	where
		tmp map f@(Normal node _ children) =
			let map' = Map.insert node f map in
			foldl' tmp map' children


compareFast (Fast a b _ _ _) (Fast a' b' _ _ _) | a==a', b==b' = CEQ
compareFast (Fast n _ _ _ _) (Fast _ n' _ _ _) | n == n' = CGT
compareFast (Fast _ p _ _ _) (Fast p' _ _ _ _) | p == p' = CLT
compareFast l@(Fast n p _ _ c) r@(Fast n' p' _ _ c') = 
	let lc = collectFast l in
	let rc = collectFast r in
	case (Map.member n' lc, Map.member n rc) of 
		(True, True) -> CEQ -- if I am my parent then I am the root of the tree
		(True, False) -> CGT
		(False, True) -> CLT
		(False, False) -> CIN -- incomparable


-- Just a tree
data Normal x = Normal { _nodeN :: x, _parentN :: x, _childrenN :: [Normal x] }
	deriving (Ord, Eq, Show)

type VisitedBitmap = UArray Word64 Bool
data NodeT = NodeT { -- _visitedInstr :: VisitedBitmap,  -- visited instructions
			  _nodeTid :: Word64, -- so that different nodes aren't equal
			  _k :: Fast NodeA  -- node of A to which we are maped by kappa
			}
	deriving (Ord, Eq, Show)

data Instruction =        InstrNew NodeA
                        | InstrJmp State
	deriving (Ord, Eq, Show)


data StateType = StateExists | StateForall
	deriving (Ord, Eq, Show)

stateNum :: StateType -> Word64 -> Word64
stateNum stateType num = 
	(num * 2) + case stateType of 
		StateExists -> 0
		StateForall -> 1

-- frame of log
data Frame = 		  FrameInstr VisitedBitmap -- represents the fact of choosing an instruction among available ones
			-- in other frame NodeA represents the _previous_ apple
			| FrameNew (Fast NodeT) (Fast NodeT) State Bool -- Bool is True if visited; NodeT is previous apple, Fast NodeT is the new node
			-- when pushing FrameJmp* we change w to a vertex in up/down set immediately
                        | FrameJmpUp (Fast NodeT) State Bool  -- Bool is True if visited
                        | FrameJmpDown (Fast NodeT) State VisitedBitmap -- bitmap of chosen vertices, we only care for bitmap size, the nodes are visited in left=>right order
-- normal flow of frames: FrameInstr describing choice of instruction,
-- 			  FrameNew/... describing action taken because of that instrucion


-- -- description of automaton
-- data GState = GState { _a :: Normal NodeA, _q0:: Map NodeA State , _instr:: Map Word64 [Instruction], _nodeas:: Map NodeA (Fast NodeA), _s:: Set State} 
-- 
-- -- instantenous description
-- data IState = IState { _q:: State, _vertices :: [Fast NodeT], _w :: Fast NodeT, _frames :: [Frame], _vertMap :: Map NodeA (Fast NodeT), _mapT :: Map NodeT (Fast NodeT), _maxnodetid :: Word64, _leftA :: Map NodeA NodeA,  _rightA :: Map NodeA NodeA,  _leftT :: Map NodeT NodeT,  _rightT :: Map NodeT NodeT, _compSuccess :: Bool }
-- compSuccess:: indicates the result of computation which is currently curried out
-- combined for Cmts.State:

data GIState = GState { _a :: Normal NodeA, _q0:: Map NodeA State , _instr:: Map Word64 [Instruction], _nodeas:: Map NodeA (Fast NodeA), _s:: Set.Set State,
			_q:: State, _vertices :: [Fast NodeT], _w :: Fast NodeT, _frames :: [Frame], _vertMap :: Map NodeA (Fast NodeT), _mapT :: Map NodeT (Fast NodeT), _maxnodetid :: Word64, _leftA :: Map NodeA NodeA,  _rightA :: Map NodeA NodeA,  _leftT :: Map NodeT NodeT,  _rightT :: Map NodeT NodeT, _compSuccess :: Bool }

-- find first free (zero) element in array (bit vector)
firstFree a =   maybe Nothing (Just . fst) $ find  ((== True) . snd)  $ assocs a

makeLenses ''NodeA
makeLenses ''Fast
makeLenses ''Normal
makeLenses ''NodeT
makeLenses ''Frame
makeLenses ''GIState
makeLenses ''State

newVisitedBitmap :: Word64 -> VisitedBitmap
newVisitedBitmap len = listArray (fromIntegral 0, len) $ replicate (fromIntegral len) False


step :: Cmts.State GIState ()
step = do
	x <- use frames
	let lastFrame = last x
	compSuccess' <- use compSuccess
	s' <- use s -- set of universal states; other are existential
	couldMoveForward <- return False
	let pushFrame f = do
		old <- use frames
		frames .= old ++ [f]
		 -- TODO: make frames an unboxed array and use the following code instead:
		 -- let addr = 1 + (last $ Data.Array.Unboxed.indices frames)
		 -- frames <- frames // (addr, f)
	let pushInstrFrame = do 
		w' <- use w
		instr' <- use instr
		q' <- use q
		let val = w'^.nodeF.k.nodeF.idA
		let len = length $ instr' Map.! val
		if (len == 0)
		then 
			compSuccess .= (Set.member q' s')
		else
			pushFrame $ FrameInstr $ newVisitedBitmap $ fromIntegral $ len
	let popFrame = do 
		old <- use frames
		frames .= init old
		
	q' <- use q
	case lastFrame of
		FrameInstr vb -> do
			-- if we are propagating a failed computation result
			--  and the state is not existential
			--  then the current state fails as well
			-- if we are in an existatial state with no children 
			--  it fails (handled in pushInstrFrame)
			if (not compSuccess' && Set.member q' s')
			then
				compSuccess .= False
			else
				case firstFree vb of 
					Nothing -> return ()
					Just iid -> do
						popFrame
						pushFrame $ FrameInstr $ vb // [(iid, True)]
						w' <- use w
						instr' <- use instr
						let pos = w'^.nodeF.k.nodeF.idA 
						let arg = (instr' Map.! pos ) !! fromIntegral iid :: Instruction
						case arg of
							InstrNew b ->  do
								w' <- use w
								nodeas' <- use nodeas
								q' <- use q
								q0' <- use q0
								rightT' <- use rightT
								leftT' <- use leftT
								maxnodetid +=1
								maxnodetid' <- use maxnodetid
								mapT' <- use mapT
								let left' = (w'^.childrenF) Map.! b -- FIXME we should be modyfying the rightmost one, not that (leftmost) one...
								let newnodet = (NodeT maxnodetid' $ nodeas' Map.! b)
								-- let newn = Fast newnodet  (w'^.nodeF) (Just left') Nothing Map.empty 
								let newn = Fast newnodet  (w'^.nodeF) () () Map.empty 
								leftT .= Map.insert newnodet (left'^.nodeF) leftT'
								mapT .= Map.insert newnodet newn mapT'
								-- let left'' = left' & right .~ Just newn  -- and write it
								rightT .= Map.insert newnodet (left'^.nodeF) rightT'
								pushFrame (FrameNew (w') newn q' False)
								q .= q0' Map.! b
								w .= newn
							InstrJmp p@(State _ levl) -> do 
								vertMap' <- use vertMap
								q' <- use q
								w' <- use w
								let lvl' = q'^.lvl
								case compareFast (vertMap' Map.! (levl) ) (vertMap' Map.! (q'^.lvl)) of
									CGT -> do
										pushFrame $ FrameJmpUp (w') q' False
										mapT' <- use mapT
										w .= mapT' Map.! (w'^.parentF)
										q .= p
									CEQ -> do -- EQ is like GT, but w is unchanged
										pushFrame $ FrameJmpUp (w') q' False
										q .= p
									CLT -> do
										let State sid _ = p 
										instr' <- use instr
										let sz = length $ instr' Map.! sid
										pushFrame $ FrameJmpDown (w') q' $ newVisitedBitmap $ fromIntegral sz
										vertMap' <- use vertMap
										w .= (vertMap' Map.! levl)
										q .= p
								-- CIN is forbidden by automaton definition
						couldMoveForward <- return True
						return ()
						where 
		FrameNew a b c v -> unless v $ do
					popFrame
					pushFrame $ FrameNew a b c True
					couldMoveForward <- return True
					pushInstrFrame
		FrameJmpUp a b v -> unless v $ do
				popFrame
				pushFrame $ FrameJmpUp a b True
				couldMoveForward <- return True
				pushInstrFrame
		FrameJmpDown a b vb -> case firstFree vb of 
			Nothing -> return ()
			Just iid -> do
				unless (iid == 0) $ do
					w' <- use w
					rightT' <- use rightT
					mapT' <- use mapT
					 -- let w'' = fromMaybe (error "no maybe") $ w'^.right
					let w'' = rightT' Map.! (w'^.nodeF)
					w .= mapT' Map.! w''
				popFrame
				pushFrame $ FrameJmpDown a b $ vb // [(iid, True)]
				couldMoveForward <- return True
				pushInstrFrame 
	
	unless couldMoveForward $ do
		case lastFrame of
			FrameInstr vb -> return ()
			FrameJmpUp w' q' _ -> do 
				w .= w'
				q .= q'
			FrameJmpDown w' q' _ -> do 
				w .= w'
				q .= q'
			FrameNew w' newn q' _ -> do
				leftT' <- use leftT
				let newnleft = (newn^.nodeF) `Map.lookup` leftT'
				when (isJust newnleft) $ do
					rightT' <- use rightT
					rightT .= (fromJust newnleft) `Map.delete` rightT'
					-- spells out: newn.left.right .= Nothing 
				w .= w'
				q .= q'
		-- end case
		
		-- remove lastFrame:
		frames' <- use frames
		frames .= init frames'

isPositive depth = (depth `mod` 2) == 1



-- normalization: representation of \forall_{A,B,C} (a->b->...->x) is ([A,B,C], [a, b, ..., x])
-- normalization: representation of \forall_{A,B} (a->b \forall_C ...->x) is ([A,B,C], [a, b, ..., x])
-- A,B,C - stored as Names
-- a, b, ..., x - stored as Types
convertType:: Type -> ([Name], [Type], Type)
convertType t@(Tall _name _type1 _type2) = result $ convertType _type2
	where 	
		result (vars, types, whole) = (vars ++ newname, types ++ [_type1, _type2], whole) 
		newname = case _name of
			"_" -> [] -- if name is "_" the quantifier binds no variables
			x -> [x]
convertType x = ([], [x], x)

-- represent formulae as trees
-- builds map (Type => Word64) during run, but it seems unnecessary (?) afterwards
-- see p. 10-11 and fig.4 in Eden Paper
-- we ASSUME tha the formula is positive, so we can know if we are at positive depth in one run, but that can be easily fixed
treePhi:: ([Name], [Type], Type) -> (Map Type Word64, Normal(NodeA, Type, Int, [Name]))
treePhi x = finish $ treePhiInternal 0 (partentZero x) (0, (Map.empty), initTree x) x
	where
		treePhiInternal:: Int -> (NodeA, Type, Int, [Name]) -> (Word64, Map Type Word64, [Normal(NodeA, Type, Int, [Name])]) -> ([Name], [Type], Type) -> (Word64, Map Type Word64, [Normal(NodeA, Type, Int, [Name])]) 
		treePhiInternal depth parent (i, m, t) (names, types', curr) = 
		-- (Word64, Map Type Int, [Normal(NodeA, Type, Int)]) is "accumulator" of state (i, map of names, list of children on the current level with their depths
				let m'    = Map.insert curr i m 
				    types = init types'
				    target= last types'
				    i'    = i+1 
				    nodeai= NodeA i in
				let cpid  = treePhiInternal (depth + 1) $ (nodeai, curr, depth, names) in 
				let (i'', m'', t') = (foldl' cpid (i', m', []) $ map convertType types) in
				(i'', m'', (Normal (nodeai, target, depth, names) parent t'):t)
		finish (a, b, c:_) = (b, c)
		nodeAZero = NodeA 0
		partentZero (n_, t_, t) = (nodeAZero, t, 0, n_)
		initTree (n_, t_, t) =
			let initTuple = (nodeAZero, t, 0, n_) in
			[Normal initTuple initTuple []]

fst4 (a, _, _, _) = a

-- convert Type to proof-searching automaton
treeAutomaton :: Type -> (Normal (NodeA, Type, Int, [Name]), (Map Word64 [Instruction], Set.Set Word64))
treeAutomaton t = (a0, cta nodea0 ((Map.empty), (Set.empty)) a0)
	where
		-- cta :: ... -> GState
		a0 = snd $ treePhi $ convertType t :: Normal (NodeA, Type, Int, [Name])
		nodea0 = case a0 of Normal (x, _, _, _) _ _ -> x
		-- we need (Normal NodeA, Map Word64 [Instruction], )
		cta :: NodeA
		       -> (Map Word64 [Instruction], Set.Set Word64)
		              -> Normal (NodeA, Type, Int, [Name])
			             -> (Map Word64 [Instruction], Set.Set Word64)

		cta parent (m, s) (Normal (nodea, type_, depth, names) parent_ children) =
			-- we build the tree top-down just like in treePhi
			-- s: set of universal states; other are existential
			let children' = withNumbers children :: [(Word64, Normal (NodeA, Type, Int, [Name]))]
			    k = length children 
			    k' = 256 
			    i' = stateToI nodea in -- ASSUME there are at most 1024 instructions per state
			let byTarget = undefined :: Map Name [(NodeA, depth)] in -- children of current node with given target
			let varNum = undefined :: Map Name Word8 in -- numbers *all variables globally*; we ASSUME that variables have different names
			let currentTarget = case last children of 
					Normal (_, Tvar x, _, _) _ _ -> x 
					_ -> error "Target not a variable"
				:: Name in
			let currentTargetNum = fromIntegral $ varNum Map.! currentTarget in
			let getVarNum (NodeA x) = undefined in -- Tvar a => varNum Map.! a in
			let (m''', s''') = if isPositive depth
				then
					--
					-- new states are numbered:q_\phi^0	|-> i'+0, 
					-- 				..., 
					-- 			   q_\phi^{k-1}	|-> i'+k-1
					-- 			   q_\phi^{k}	|-> i'+k'
					-- 			   p_\phi^{x}	|-> i'+k'+number of x; note that - different than in the paper - we do not have predicates
					-- 			   			assume that always k<k'
					-- and then i is incremented by k+#predicated
					let m'	= foldl' (\m_ (num, Normal (part, _, _, _) _ _) -> Map.insert (i'+num) [InstrNew $ part]  m_) m $ children' in-- q_\phi^i: new \psi_{i+1} 
					let m''	= (flip $ Map.insert (i'+k')) m' $ catMaybes $ map (\(nodea, depth) -> if isPositive depth then Just $ InstrJmp $ State ((stateToI nodea) + k' + currentTargetNum) nodea else Nothing) $ Map.findWithDefault [] currentTarget byTarget :: Map Word64 [Instruction] in-- q_\phi^k: jmp p_\theta^y where y=currentTarget 
					(foldl' (\m__ name -> 
								foldl' (\m_ (zeta, zdepth) -> 
										if not (isPositive zdepth) && nodeaGE zeta nodea
										then Map.insert (i'+1+getVarNum zeta) [InstrJmp $ State (stateToI zeta + 1) zeta] m_
										else m_ )
								m__ (byTarget Map.! name)
							) m'' names, s)
					
				else --phi is negaive
					-- let 	l  = parentAssumptionNumber nodea parent_ :: Int 
					let 	l  = undefined :: Int 
						m' = Map.insert (i'+k') [InstrJmp $ State (k' + (stateToI $ fst4 parent_)) $ fst4 parent_] m
						m''= Map.insert (i'+k'+1) [InstrJmp $ State (i' + 1) nodea] m'
						s' = Set.insert (i'+1) s in -- THE universal state
					(foldl' (\m_ (num, Normal (part, _, _, _) _ _) ->  Map.insert (i'+num) [InstrJmp $ State (stateToI part) part]  m_) m'' children', s')
			in
			-- now process children
			let cta' = cta nodea in
			foldl' cta' (m''', s''') children
				
			
--		isPositive depth = (depth `mod` 2) == 1
		withNumbers  = zip [0..] -- numbers start at zero
		stateToI (NodeA id) = id * 1024
		parentAssumptionNumber nodea (Normal (_, _, _ , _) _ pchldren) = fst $ head $ filter (\(_,x) -> x == nodea) $ withNumbers pchldren
		nodeaGEfind  xi phi  n@(Normal (nodea, type_, depth, names) parent_ children) = -- find xi and look for phi in it's children
			if nodea == xi 
				then nodeaGEcheck phi n
				else Prelude.or $ map (nodeaGEfind xi phi) children
		nodeaGEcheck phi  n@(Normal (nodea, type_, depth, names) parent_ children) = -- find xi and look for phi in it's children
			nodea == phi || (Prelude.or $ map (nodeaGEcheck phi) children)
		nodeaGE xi phi = nodeaGEfind xi phi a0
	

stripA4 :: Normal (NodeA, Type, Int, [Name]) -> Normal NodeA
stripA4 (Normal node parent children) = Normal (fst4 node) (fst4 parent) (map stripA4 children)

pseudoNormal2Fast :: Normal NodeA -> Map NodeA (Fast NodeA)
pseudoNormal2Fast n = internal Map.empty n
	where
	 internal m (Normal node parent children) = 
	 	-- let node' = fst4 node in 
		-- let parent' = fst4 parent in 
		Map.insert node (Fast node parent () () undefined) $ foldl' internal m children 


typeToGSTate t =
	let (tt, (instr, set)) = treeAutomaton t in -- treeAutomaton :: Type -> (Normal (NodeA, Type, Int, [Name]), (Map Word64 [Instruction], Set.Set Word64))
	let 	(Normal (root, _, _, _) _ _) = tt -- root:: NodeA
		a = stripA4 tt 
		q = State 0 root
		fastroot = nodeas Map.! root
		nodet = NodeT 0 fastroot 
		fastnodet = Fast nodet nodet () () me
		maxnodetid = 0
		me = Map.empty
		q0 = Map.singleton root q
		nodeas = pseudoNormal2Fast a
		vertmap = Map.singleton root fastnodet
		mapt = Map.singleton nodet fastnodet
		s = undefined :: Set.Set State
	in
	GState  a q0 instr nodeas s q [fastnodet] fastnodet [] vertmap mapt maxnodetid me me me me True
